/* Copyright 2018 R. Thomas
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include "pyAMI.hpp"
#include "version.h"
#include "AMI/AMI.hpp"
//#include "AMI/version.h"
#include <sstream>

py::module ami_module("pyAMI", "AMI Python API");

using namespace AMI;

PYBIND11_MODULE(pyAMI, ami_module) {

  ami_module.attr("__version__")   = py::str(AMI_VERSION);
  ami_module.attr("__tag__")       = py::str(AMI_TAG);
  ami_module.attr("__is_tagged__") = py::bool_(AMI_TAGGED);

  py::class_<IntentFilter> intent_filter(ami_module, "IntentFilter", "Android IntentFilter representation");

  py::class_<IntentFilter::data_t> (intent_filter, "IntentFilter.data_t", "Data associated with an IntentFilter")
    .def_readonly("scheme",
        &IntentFilter::data_t::scheme)

    .def_readonly("host",
        &IntentFilter::data_t::host)

    .def_readonly("path",
        &IntentFilter::data_t::path)

    .def_readonly("path_pattern",
        &IntentFilter::data_t::path_pattern)

    .def_readonly("path_prefix",
        &IntentFilter::data_t::path_prefix)

    .def_readonly("mime_type",
        &IntentFilter::data_t::mime_type)

    .def("__str__",
        [] (const IntentFilter::data_t& data) {
          std::ostringstream oss;
          oss << data;
          return oss.str();
        });


  intent_filter
    .def_property_readonly("actions",
        &IntentFilter::actions)

    .def_property_readonly("categories",
        &IntentFilter::categories)

    .def_property_readonly("data",
        &IntentFilter::data)

    .def("__str__",
        [] (const IntentFilter& filter) {
          std::ostringstream oss;
          oss << filter;
          return oss.str();
        });

  py::class_<Service>(ami_module, "Service", "Android Service representation")
    .def_property_readonly("cls",
        &Service::cls)

    .def_property_readonly("is_exported",
        &Service::is_exported)

    .def_property_readonly("is_enabled",
        &Service::is_enabled)

    .def_property_readonly("has_permission",
        &Service::has_permission)

    .def_property_readonly("permission",
        &Service::permission)

    .def_property_readonly("filters",
        &Service::filters)

    .def("__str__",
        [] (const Service& service) {
          std::ostringstream oss;
          oss << service;
          return oss.str();
        });


  py::class_<Permission>(ami_module, "Permission", "Android Permission representation")

    .def_property_readonly("fullname",
        &Permission::fullname)

    .def("__str__",
        [] (const Permission& perm) {
          std::ostringstream oss;
          oss << perm;
          return oss.str();
        });

  py::class_<Activity>(ami_module, "Activity", "Android Activity representation")
    .def_property_readonly("cls",
        &Activity::cls)

    .def("__str__",
        [] (const Activity& activity) {
          std::ostringstream oss;
          oss << activity;
          return oss.str();
        });


  py::class_<BroadcastReceiver>(ami_module, "BroadcastReceiver", "Android BroadcastReceiver representation")
    .def_property_readonly("cls",
        &BroadcastReceiver::cls)

    .def_property_readonly("is_exported",
        &BroadcastReceiver::is_exported)

    .def_property_readonly("is_direct_boot_aware",
        &BroadcastReceiver::is_direct_boot_aware)

    .def_property_readonly("is_enabled",
        &BroadcastReceiver::is_enabled)

    .def_property_readonly("has_permission",
        &BroadcastReceiver::has_permission)

    .def_property_readonly("permission",
        &BroadcastReceiver::permission)

    .def_property_readonly("filters",
        &BroadcastReceiver::filters)

    .def("__str__",
        [] (const BroadcastReceiver& br) {
          std::ostringstream oss;
          oss << br;
          return oss.str();
        });

  py::class_<Provider>(ami_module, "Provider", "Android Provider representation")
    .def_property_readonly("cls",
        &Provider::cls)

    .def_property_readonly("is_exported",
        &Provider::is_exported)

    .def_property_readonly("is_enabled",
        &Provider::is_enabled)

    .def_property_readonly("is_direct_boot_aware",
        &Provider::is_direct_boot_aware)

    .def_property_readonly("authorities",
        &Provider::authorities)

    .def("__str__",
        [] (const Provider& obj) {
          std::ostringstream oss;
          oss << obj;
          return oss.str();
        });



  py::class_<Application>(ami_module, "Application", "Android Application representation")
    .def_property_readonly("cls",
        &Application::cls)

    .def_property_readonly("process",
        &Application::process)

    .def_property_readonly("backup_agent",
        &Application::backup_agent)

    .def_property_readonly("is_debuggable",
        &Application::is_debuggable)

    .def("__str__",
        [] (const Application& app) {
          std::ostringstream oss;
          oss << app;
          return oss.str();
        });

  py::class_<AXML>(ami_module, "AXML", "AXML Interface")

    .def_static("open",
        [] (py::bytes bytes) -> py::object {

          std::string raw_str = bytes;
          std::vector<uint8_t> raw = {
            std::make_move_iterator(std::begin(raw_str)),
            std::make_move_iterator(std::end(raw_str))
          };

          AXML axml = AXML::open(std::move(raw));
          if (axml) {
            return py::cast(axml);
          } else {
            return py::none();
          }
        })
    .def_static("open",
        [] (const std::string& path) -> py::object {
          AXML axml = AXML::open(path);
          if (axml) {
            return py::cast(axml);
          } else {
            return py::none();
          }
        })

    .def("__bool__",
        [] (const AXML& axml) {
          return static_cast<bool>(axml);
        })

    .def_property_readonly("xml",
        static_cast<XML::Root&(AXML::*)(void)>(&AXML::xml),
        py::return_value_policy::reference)

    .def_property_readonly("string_pool_size",
        &AXML::string_pool_size)

    .def_property_readonly("string_pool",
        &AXML::string_pool)

    .def("__str__",
        [] (const AXML& obj) {
          std::ostringstream ss;
          ss << obj;
          return ss.str();
        });

  py::class_<Manifest, AXML>(ami_module, "Manifest", "Android Manifest representation")
    .def_static("open",
        [] (const std::string& path) -> py::object {
          Manifest manifest = Manifest::open(path);
          if (manifest) {
            return py::cast(manifest);
          } else {
            return py::none();
          }
        })

    .def_property_readonly("package_name",
        &Manifest::package_name)

    .def_property_readonly("version_name",
        &Manifest::version_name)

    .def_property_readonly("version_code",
        &Manifest::version_code)

    .def_property_readonly("permissions",
        &Manifest::permissions)

    .def_property_readonly("services",
        &Manifest::services)

    .def_property_readonly("activities",
        &Manifest::activities)

    .def_property_readonly("receivers",
        &Manifest::receivers)

    .def_property_readonly("providers",
        &Manifest::providers)

    .def_property_readonly("application",
        [] (const Manifest& m) -> py::object {
          Application app = m.application();
          if (app) {
            return py::cast(app);
          }
          return py::none();
        })

    .def("__bool__",
        [] (const Manifest& manifest) {
          return static_cast<bool>(manifest);
        })

    .def_property_readonly("xml",
        static_cast<XML::Manifest&(Manifest::*)(void)>(&Manifest::xml),
        py::return_value_policy::reference)

    .def("__str__",
        [] (const Manifest& m) {
          std::ostringstream ss;
          ss << m;
          return ss.str();
        });

  py::module xml_module = ami_module.def_submodule("XML", "XML API");
  init_xml(xml_module);
}
