/* Copyright 2018 R. Thomas
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#ifndef AMI_ANDROID_MANIFEST_H_
#define AMI_ANDROID_MANIFEST_H_
#include <memory>
#include <string>
#include <vector>

#include <AMI/visibility.h>
#include <AMI/XML.hpp>


namespace AMI {
namespace details {
class AndroidManifestImpl;
}

class AMI_API IntentFilter {
  public:
  struct data_t {
    std::string scheme;
    std::string host;
    std::string port;
    std::string path;
    std::string path_pattern;
    std::string path_prefix;
    std::string mime_type;

    AMI_API friend std::ostream& operator<<(std::ostream& os, const data_t& m);
  };
  using action_list_t   = std::vector<std::string>;
  using data_list_t     = std::vector<data_t>;
  using category_list_t = std::vector<std::string>;

  public:
  IntentFilter(void) = delete;
  IntentFilter(const XML::Element& e);

  IntentFilter(const IntentFilter&);
  IntentFilter& operator=(const IntentFilter&);

  bool is_valid(void) const;
  operator bool() const;

  action_list_t   actions() const;
  category_list_t categories() const;
  data_list_t     data() const;

  AMI_API friend std::ostream& operator<<(std::ostream& os, const IntentFilter& m);

  private:
  XML::Element* element_{nullptr};
};



class AMI_API Permission {
  public:
  Permission(void) = delete;
  Permission(const XML::Element& e);

  Permission(const Permission&);
  Permission& operator=(const Permission&);

  std::string fullname(void) const;

  AMI_API friend std::ostream& operator<<(std::ostream& os, const Permission& m);

  private:
  XML::Element* element_{nullptr};
};

class AMI_API Activity {
  public:
  Activity(void) = delete;
  Activity(const XML::Element& e);

  Activity(const Activity&);
  Activity& operator=(const Activity&);

  std::string cls(void) const;

  AMI_API friend std::ostream& operator<<(std::ostream& os, const Activity& m);

  private:
  XML::Element* element_{nullptr};
};


class AMI_API Service {
  public:
  using filters_t = std::vector<IntentFilter>;

  public:
  Service(void) = delete;
  Service(const XML::Element& e);

  Service(const Service&);
  Service& operator=(const Service&);

  std::string cls(void) const;

  bool is_valid(void) const;
  operator bool() const;

  bool is_exported(void) const;
  bool is_enabled(void) const;
  bool is_isolated(void) const;
  bool has_permission(void) const;
  std::string permission(void) const;

  filters_t filters() const;

  AMI_API friend std::ostream& operator<<(std::ostream& os, const Service& m);

  private:
  XML::Element* element_{nullptr};
};


class AMI_API Application {

  public:
  Application(void);
  Application(const XML::Element& e);

  Application(const Application&);
  Application& operator=(const Application&);

  std::string cls(void) const;
  bool is_debuggable(void) const;
  std::string process(void) const;
  std::string backup_agent(void) const;

  operator bool() const;

  AMI_API friend std::ostream& operator<<(std::ostream& os, const Application& m);

  private:
  XML::Element* element_{nullptr};
};


class AMI_API BroadcastReceiver {
  public:
  using filters_t = std::vector<IntentFilter>;

  public:
  BroadcastReceiver(void) = delete;
  BroadcastReceiver(const XML::Element& e);

  BroadcastReceiver(const BroadcastReceiver&);
  BroadcastReceiver& operator=(const BroadcastReceiver&);

  bool is_valid(void) const;
  operator bool() const;

  std::string cls(void) const;
  bool is_exported(void) const;
  bool is_direct_boot_aware(void) const;
  bool is_enabled(void) const;
  bool has_permission(void) const;
  std::string permission(void) const;
  filters_t filters() const;

  AMI_API friend std::ostream& operator<<(std::ostream& os, const BroadcastReceiver& m);

  private:
  XML::Element* element_{nullptr};
};


class AMI_API Provider {
  public:
  Provider(void) = delete;
  Provider(const XML::Element& e);

  Provider(const Provider&);
  Provider& operator=(const Provider&);

  bool is_valid(void) const;
  operator bool() const;

  std::string cls(void) const;

  bool is_exported(void) const;
  bool is_enabled(void) const;
  bool is_direct_boot_aware() const;
  std::string authorities() const;

  AMI_API friend std::ostream& operator<<(std::ostream& os, const Provider& m);

  private:
  XML::Element* element_{nullptr};
};

class AMI_API AXML {
  public:
  using blob_t = std::vector<uint8_t>;
  using string_pool_t = std::vector<std::string>;

  AXML(void) = delete;
  AXML(const AXML& other);
  AXML& operator=(const AXML& other);

  static AXML open(const std::string& path);
  static AXML open(const char* buffer, size_t size);
  static AXML open(const blob_t& blob);
  static AXML open(blob_t&& blob);

  XML::Root&       xml(void);
  const XML::Root& xml(void) const;

  operator bool() const;

  AMI_API friend std::ostream& operator<<(std::ostream& os, const AXML& m);

  ~AXML(void);

  size_t string_pool_size(void) const;

  const string_pool_t& string_pool(void) const;

  protected:
  AXML(const blob_t& blob);
  AXML(std::unique_ptr<details::AndroidManifestImpl>&& impl);
  std::unique_ptr<details::AndroidManifestImpl> pimpl_;
};


class AMI_API Manifest : public AXML {
  public:
  using permission_list_t = std::vector<Permission>;
  using services_list_t   = std::vector<Service>;
  using activities_list_t = std::vector<Activity>;
  using receivers_list_t  = std::vector<BroadcastReceiver>;
  using providers_list_t  = std::vector<Provider>;

  public:
  Manifest(void) = delete;
  Manifest(const Manifest& other);
  Manifest& operator=(const Manifest& other);

  static Manifest open(const std::string& path);
  static Manifest open(const char* buffer, size_t size);
  static Manifest open(const AXML::blob_t& blob);
  static Manifest open(AXML::blob_t&& blob);

  std::string package_name(void) const;
  std::string version_name(void) const;
  size_t      version_code(void) const;

  operator bool() const;

  permission_list_t permissions(void) const;
  services_list_t   services(void) const;
  activities_list_t activities(void) const;
  receivers_list_t  receivers(void) const;
  providers_list_t  providers(void) const;
  Application       application(void) const;

  XML::Manifest&       xml(void);
  const XML::Manifest& xml(void) const;

  AMI_API friend std::ostream& operator<<(std::ostream& os, const Manifest& m);

  ~Manifest(void);

  private:
  Manifest(const blob_t& blob);
  Manifest(std::unique_ptr<details::AndroidManifestImpl>&& impl);
};



} // Namespace AMI

#endif
