/* Copyright 2018 R. Thomas
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "AMI/AMI.hpp"
#include "AndroidManifestImpl.hpp"
#include "macros.h"

namespace AMI {

BroadcastReceiver::BroadcastReceiver(const XML::Element& e) :
  element_{const_cast<XML::Element*>(&e)}
{}

BroadcastReceiver::BroadcastReceiver(const BroadcastReceiver&) = default;
BroadcastReceiver& BroadcastReceiver::operator=(const BroadcastReceiver&) = default;


std::string BroadcastReceiver::cls(void) const {
  CHECK_NE(this->element_, nullptr);
  const XML::Attribute* attr = (*this->element_)["android:name"];
  if (attr == nullptr) {
    return "";
  }
  return (*this->element_)["android:name"]->svalue();
}

bool BroadcastReceiver::is_valid(void) const {
  return this->element_ != nullptr;
}

BroadcastReceiver::operator bool() const {
  return this->is_valid();
}

bool BroadcastReceiver::is_exported(void) const {
  if (not this->is_valid()) {
    return true;
  }
  const XML::Attribute* attr = (*this->element_)["android:exported"];
  if (attr == nullptr) { /* Default value */
    return true;
  }
  return static_cast<bool>(attr->ivalue());
}

bool BroadcastReceiver::is_direct_boot_aware(void) const {
  if (not this->is_valid()) {
    return false;
  }
  const XML::Attribute* attr = (*this->element_)["android:directBootAware"];
  if (attr == nullptr) { /* Default value */
    return false;
  }
  return static_cast<bool>(attr->ivalue());
}

bool BroadcastReceiver::is_enabled(void) const {
  if (not this->is_valid()) {
    return true;
  }
  const XML::Attribute* attr = (*this->element_)["android:enabled"];
  if (attr == nullptr) { /* Default value */
    return true;
  }
  return static_cast<bool>(attr->ivalue());
}

bool BroadcastReceiver::has_permission(void) const {
  if (not this->is_valid()) {
    return false;
  }
  const XML::Attribute* attr = (*this->element_)["android:permission"];
  return attr != nullptr;
}

std::string BroadcastReceiver::permission(void) const {
  if (this->has_permission()) {
    const XML::Attribute* attr = (*this->element_)["android:permission"];
    return attr->svalue();
  }
  return "";
}

BroadcastReceiver::filters_t BroadcastReceiver::filters() const {
  if (not this->is_valid()) {
    return {};
  }
  BroadcastReceiver::filters_t filters;

  XML::const_elements_t intent_filters = this->element_->find_elements("intent-filter");

  for  (const XML::Element* e : intent_filters) {
    filters.emplace_back(*e);
  }
  return filters;
}


std::ostream& operator<<(std::ostream& os, const BroadcastReceiver& m) {
  os << m.cls();
  return os;
}


}
