/* Copyright 2018 R. Thomas
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "AMI/AMI.hpp"
#include "AndroidManifestImpl.hpp"
#include "macros.h"
#include <sstream>

#include <spdlog/spdlog.h>
#include <spdlog/fmt/ostr.h>

namespace AMI {

IntentFilter::IntentFilter(const XML::Element& e) :
  element_{const_cast<XML::Element*>(&e)}
{}

IntentFilter::IntentFilter(const IntentFilter&) = default;
IntentFilter& IntentFilter::operator=(const IntentFilter&) = default;

IntentFilter::action_list_t IntentFilter::actions() const {
  if (not this->is_valid()) {
    return {};
  }
  IntentFilter::action_list_t actions;
  XML::const_elements_t actions_elements = this->element_->find_elements("action");

  for (const XML::Element* a : actions_elements) {
    const XML::Attribute* attr = a->get("android:name");
    if (attr == nullptr) {
      continue;
    }
    std::string action_str = attr->svalue();
    actions.push_back(action_str);
  }
  return actions;
}

IntentFilter::category_list_t IntentFilter::categories() const {
  if (not this->is_valid()) {
    return {};
  }
  IntentFilter::category_list_t categories;
  XML::const_elements_t category_elements = this->element_->find_elements("category");

  for (const XML::Element* a : category_elements) {
    const XML::Attribute* attr = a->get("android:name");
    if (attr == nullptr) {
      continue;
    }
    std::string action_str = attr->svalue();
    categories.push_back(action_str);
  }
  return categories;
}

IntentFilter::data_list_t IntentFilter::data() const {
  if (not this->is_valid()) {
    return {};
  }
  IntentFilter::data_list_t data_list;
  XML::const_elements_t data_elements = this->element_->find_elements("data");

  for (const XML::Element* a : data_elements) {
    data_t data;
    const XML::Attribute* scheme       = a->get("android:scheme");
    const XML::Attribute* host         = a->get("android:host");
    const XML::Attribute* port         = a->get("android:port");
    const XML::Attribute* path         = a->get("android:path");
    const XML::Attribute* path_prefix  = a->get("android:pathPrefix");
    const XML::Attribute* path_pattern = a->get("android:pathPattern");
    const XML::Attribute* mime_type    = a->get("android:mimeType");

    if (scheme != nullptr) data.scheme = scheme->svalue();
    if (port != nullptr) data.port = port->svalue();
    if (host != nullptr) data.host = host->svalue();
    if (path != nullptr) data.path = path->svalue();
    if (path_prefix != nullptr)  data.path_prefix  = path_prefix->svalue();
    if (path_pattern != nullptr) data.path_pattern = path_pattern->svalue();
    if (mime_type != nullptr)    data.mime_type    = mime_type->svalue();

    data_list.push_back(std::move(data));
  }
  return data_list;
}

bool IntentFilter::is_valid(void) const {
  return this->element_ != nullptr;
}

IntentFilter::operator bool() const {
  return this->is_valid();
}

std::ostream& operator<<(std::ostream& os, const IntentFilter& m) {
  std::string out = "";
  if (m.actions().size() > 0) {
    out += fmt::format("Actions ({:02d})\n", m.actions().size());
    for (const std::string& action : m.actions()) {
      out += "    " + action + "\n";
    }
  }

  if (m.categories().size() > 0) {
    out += fmt::format("Categories ({:02d})\n", m.categories().size());
    for (const std::string& category : m.categories()) {
      out += "    " + category + "\n";
    }
  }

  if (m.data().size() > 0) {
    out += fmt::format("Data ({:02d})\n", m.data().size());
    for (const IntentFilter::data_t& data : m.data()) {
      std::ostringstream oss;
      oss << data;
      out += "    " + oss.str() + "\n";
    }
  }

  os << out;
  return os;
}


std::ostream& operator<<(std::ostream& os, const IntentFilter::data_t& m) {
  std::string out = "";
  if (not m.scheme.empty()) {
    out += m.scheme + "://";
  }

  if (not m.host.empty()) {
    out += m.host;
  }

  if (not m.port.empty()) {
    out += ":" + m.port;
  }

  if (not m.path.empty()) {
    out += "/" + m.path;
  }

  if (not m.path_pattern.empty()) {
    out += "/" + m.path_pattern;
  }

  if (not m.path_prefix.empty()) {
    out += "/" + m.path_prefix;
  }
  os << out;
  return os;
}


}
