/* Copyright 2018 R. Thomas
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include "AMI/AMI.hpp"
#include "AndroidManifestImpl.hpp"
#include "macros.h"

#include <algorithm>

namespace AMI {

Manifest::~Manifest(void) = default;

Manifest Manifest::open(const std::string& path) {
  std::unique_ptr<details::AndroidManifestImpl> impl{details::AndroidManifestImpl::open(path)};
  return std::move(impl);
}

Manifest Manifest::open(const char* raw, size_t size) {
  return Manifest::blob_t{raw, raw + size};
}

Manifest::Manifest(const Manifest::blob_t& blob) :
  AXML::AXML{blob}
{}

Manifest::Manifest(const Manifest& other) :
  AXML::AXML{other}
{}


std::string Manifest::package_name(void) const {
  CHECK(static_cast<bool>(this->pimpl_));
  const XML::Node::children_t& children = this->pimpl_->root().children();
  if (children.size() != 1) {
    return "";
  }
  const XML::Node* manifest_node = this->pimpl_->root().children()[0];

  XML::const_elements_t manifest_element = manifest_node->find_elements("manifest", /* only_one */ true);
  CHECK(manifest_element.size() == 1);

  const XML::Element* manifest = manifest_element.back();
  XML::const_attributes_t attributes = manifest->find_attributes("package", /* only_one */ true);
  if (attributes.size() == 0) {
    return "";
  }
  const XML::Attribute* package = attributes[0];
  CHECK(package != nullptr);

  // with string type
  switch (package->type()) {
    case XML::Attribute::TYPES::STRING:
      {
        return package->svalue();
        break;
      }

    default:
      {
        return "";
      }
  }

}

std::string Manifest::version_name(void) const {
  CHECK(static_cast<bool>(this->pimpl_));

  const XML::Node::children_t& children = this->pimpl_->root().children();
  CHECK(children.size() == 1);

  const XML::Node* manifest_node = this->pimpl_->root().children()[0];
  CHECK(manifest_node != nullptr);

  XML::const_elements_t manifest_element = manifest_node->find_elements("manifest", /* only_one */ true);
  CHECK(manifest_element.size() == 1);

  const XML::Element* manifest = manifest_element.back();
  CHECK(manifest != nullptr);

  XML::const_attributes_t attributes = manifest->find_attributes("android:versionName", /* only_one */ true);
  if (attributes.size() == 0) {
    return "";
  }

  // with string type
  switch (attributes[0]->type()) {
    case XML::Attribute::TYPES::STRING:
      {
        return attributes[0]->svalue();
        break;
      }

    default:
      {
        return "";
      }
  }
}

size_t Manifest::version_code(void) const {
  CHECK(static_cast<bool>(this->pimpl_));

  const XML::Node::children_t& children = this->pimpl_->root().children();
  CHECK(children.size() == 1);

  const XML::Node* manifest_node = this->pimpl_->root().children()[0];
  CHECK(manifest_node != nullptr);

  XML::const_elements_t manifest_element = manifest_node->find_elements("manifest", /* only_one */ true);
  CHECK(manifest_element.size() == 1);

  const XML::Element* manifest = manifest_element.back();
  CHECK(manifest != nullptr);

  XML::const_attributes_t attributes = manifest->find_attributes("android:versionCode", /* only_one */ true);
  if (attributes.size() == 0) {
    return 0;
  }

  // Only one version defined
  if (attributes.size() == 0) {
    return 0;
  }

  // with string type
  return attributes[0]->ivalue();
}


Manifest::permission_list_t Manifest::permissions(void) const {
  const XML::Node::children_t& children = this->pimpl_->root().children();
  CHECK(children.size() == 1);

  const XML::Node* manifest_node = this->pimpl_->root().children()[0];
  CHECK(manifest_node != nullptr);

  XML::const_elements_t permissions = manifest_node->find_elements("uses-permission");
  Manifest::permission_list_t permissions_obj;
  permissions_obj.reserve(permissions.size());

  std::transform(
      std::begin(permissions),
      std::end(permissions),
      std::back_inserter(permissions_obj),
      [] (const XML::Element* e) -> const XML::Element& {
        return *e;
      });
  return permissions_obj;
}


Manifest::services_list_t Manifest::services(void) const {
  const XML::Node::children_t& children = this->pimpl_->root().children();
  CHECK(children.size() == 1);

  const XML::Node* manifest_node = this->pimpl_->root().children()[0];
  CHECK(manifest_node != nullptr);

  XML::const_elements_t services = manifest_node->find_elements("service");
  Manifest::services_list_t services_obj;
  services_obj.reserve(services.size());

  std::transform(
      std::begin(services),
      std::end(services),
      std::back_inserter(services_obj),
      [] (const XML::Element* e) -> const XML::Element& {
        return *e;
      });
  return services_obj;

}

Manifest::activities_list_t Manifest::activities(void) const {
  const XML::Node::children_t& children = this->pimpl_->root().children();
  CHECK(children.size() == 1);

  const XML::Node* manifest_node = this->pimpl_->root().children()[0];
  CHECK(manifest_node != nullptr);

  XML::const_elements_t activities = manifest_node->find_elements("activity");
  Manifest::activities_list_t activity_obj;
  activity_obj.reserve(activities.size());

  std::transform(
      std::begin(activities),
      std::end(activities),
      std::back_inserter(activity_obj),
      [] (const XML::Element* e) -> const XML::Element& {
        return *e;
      });
  return activity_obj;
}

Manifest::receivers_list_t Manifest::receivers(void) const {
  const XML::Node::children_t& children = this->pimpl_->root().children();
  CHECK(children.size() == 1);

  const XML::Node* manifest_node = this->pimpl_->root().children()[0];
  CHECK(manifest_node != nullptr);

  XML::const_elements_t br = manifest_node->find_elements("receiver");
  Manifest::receivers_list_t br_obj;
  br_obj.reserve(br.size());

  std::transform(
      std::begin(br),
      std::end(br),
      std::back_inserter(br_obj),
      [] (const XML::Element* e) -> const XML::Element& {
        return *e;
      });
  return br_obj;
}


Manifest::providers_list_t Manifest::providers(void) const {
  static const char* KEYWORD = "provider";
  const XML::Node::children_t& children = this->pimpl_->root().children();
  CHECK(children.size() == 1);

  const XML::Node* manifest_node = this->pimpl_->root().children()[0];
  CHECK(manifest_node != nullptr);

  XML::const_elements_t elements = manifest_node->find_elements(KEYWORD);
  Manifest::providers_list_t objs;
  objs.reserve(elements.size());

  std::transform(
      std::begin(elements),
      std::end(elements),
      std::back_inserter(objs),
      [] (const XML::Element* e) -> const XML::Element& {
        return *e;
      });
  return objs;
}


Application Manifest::application(void) const {
  const XML::Node::children_t& children = this->pimpl_->root().children();
  CHECK(children.size() == 1);

  const XML::Node* manifest_node = this->pimpl_->root().children()[0];
  CHECK(manifest_node != nullptr);

  XML::const_elements_t br = manifest_node->find_elements("application", true);
  if (br.size() > 0) {
    return *br[0];
  }
  return {};

}


Manifest& Manifest::operator=(const Manifest& other) {
  if (*this and this != &other) {
    this->pimpl_.reset(new details::AndroidManifestImpl{*other.pimpl_});
  }
  return *this;
}

Manifest::Manifest(std::unique_ptr<details::AndroidManifestImpl>&& impl) :
  AXML::AXML{std::move(impl)}
{}


XML::Manifest& Manifest::xml(void) {
  return const_cast<XML::Manifest&>(static_cast<const Manifest*>(this)->xml());
}

const XML::Manifest& Manifest::xml(void) const {
  const XML::Node::children_t& children = this->pimpl_->root().children();
  CHECK(children.size() == 1);
  return reinterpret_cast<XML::Manifest&>(*children[0]);
}

Manifest::operator bool() const {
  return this->pimpl_ != nullptr and
    static_cast<bool>(*this->pimpl_) and
    this->pimpl_->root().children().size() == 1;
}

std::ostream& operator<<(std::ostream& os, const Manifest& m) {
  static const std::string SPACE = "   ";
  Manifest::activities_list_t activities  = m.activities();
  Manifest::services_list_t services      = m.services();
  Manifest::permission_list_t permissions = m.permissions();
  Manifest::receivers_list_t receivers    = m.receivers();
  os << m.package_name() << " - " << m.version_name() << std::endl;

  os << "Permissions (" << std::dec << permissions.size() << "):" << std::endl;
  for (const Permission& p : permissions) {
    os << SPACE << p << std::endl;
  }

  os << std::endl;

  os << "Activities (" << std::dec << activities.size() << "):" << std::endl;
  for (const Activity& a : activities) {
    os << SPACE << a << std::endl;
  }

  os << std::endl;

  os << "Services (" << std::dec << services.size() << "):" << std::endl;
  for (const Service& s : services) {
    os << SPACE << s << std::endl;
  }

  os << std::endl;


  os << "Broadcast Receivers (" << std::dec << receivers.size() << "):" << std::endl;
  for (const BroadcastReceiver& br : receivers) {
    os << SPACE << br << std::endl;
  }


  return os;
}

} // Namespace AMI
