/* Copyright 2018 R. Thomas
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "AMI/AMI.hpp"
#include "AndroidManifestImpl.hpp"
#include "macros.h"

namespace AMI {

Service::Service(const XML::Element& e) :
  element_{const_cast<XML::Element*>(&e)}
{}

Service::Service(const Service&) = default;
Service& Service::operator=(const Service&) = default;

bool Service::is_valid(void) const {
  return this->element_ != nullptr;
}
Service::operator bool() const {
  return this->is_valid();
}


std::string Service::cls(void) const {
  if (not this->is_valid()) {
    return "";
  }
  const XML::Attribute* attr = (*this->element_)["android:name"];
  if (attr == nullptr) {
    return "";
  }
  return attr->svalue();
}

bool Service::is_exported(void) const {
  if (not this->is_valid()) {
    return "";
  }
  const XML::Attribute* attr = (*this->element_)["android:exported"];
  if (attr == nullptr) {
    return true; /* Default value */
  }
  return static_cast<bool>(attr->ivalue());
}

bool Service::is_enabled(void) const {
  if (not this->is_valid()) {
    return false;
  }
  const XML::Attribute* attr = (*this->element_)["android:enabled"];
  if (attr == nullptr) {
    return true; /* Default value */
  }
  return static_cast<bool>(attr->ivalue());
}

bool Service::is_isolated(void) const {
  if (not this->is_valid()) {
    return false;
  }
  const XML::Attribute* attr = (*this->element_)["android:isolatedProcess"];
  if (attr == nullptr) {
    return false; /* Default value */
  }
  return static_cast<bool>(attr->ivalue());
}

bool Service::has_permission(void) const {
  if (not this->is_valid()) {
    return false;
  }

  const XML::Attribute* attr = (*this->element_)["android:permission"];
  return attr != nullptr;
}
std::string Service::permission(void) const {
  if (this->has_permission()) {
    const XML::Attribute* attr = (*this->element_)["android:permission"];
    return attr->svalue();
  }
  return "";
}

Service::filters_t Service::filters() const {
  if (not this->is_valid()) {
    return {};
  }
  BroadcastReceiver::filters_t filters;

  XML::const_elements_t intent_filters = this->element_->find_elements("intent-filter");

  for  (const XML::Element* e : intent_filters) {
    filters.emplace_back(*e);
  }
  return filters;
}

std::ostream& operator<<(std::ostream& os, const Service& m) {
  static constexpr const char* SPACE = "   ";
  os << m.cls();
  if (m.is_exported()) {
    os << " - EXPORTED";
  }

  //auto&& actions = m.actions();
  //if (actions.size() > 0) {
  //  os << std::endl;
  //  for (const std::string& action : m.actions()) {
  //    os << SPACE << action << std::endl;
  //  }
  //}
  return os;
}


}
